/*
 * kgraft_patch_bsc1157770
 *
 * Fix for bsc#1157770
 *
 *  Upstream commit:
 *  b23e5844dfe7 ("xen/pv: Fix a boot up hang revealed by int3 self test")
 *
 *  SLE12-SP1 commit:
 *  not affected
 *
 *  SLE12-SP2 commit
 *  b6986af08b78c961b18d20fa8541c646d925f2d5
 *
 *  SLE12-SP3 commit:
 *  7396f6d4f4c68bf73fb5e7f23ebf8586a2563b76
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  bf509214b68459f202946a746abd8f4bf267577d
 *
 *
 *  Copyright (c) 2019 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Background + rationale
 * ----------------------
 * For XEN PV guests, the hypervisor wraps the guest's exception entry
 * handlers with its own ones. To this end, the PV guest announces its
 * entry handler addresses via certain hypercalls to the hypervisor, see
 * e.g. xen_load_idt().
 *
 * Usually, xen_load_idt() & Co would just pass the address of the
 * native handler as set in the IDT table. But it special cases on
 * those ones which are known to use ISTs in the native case and
 * substitutes non-IST variants for those.
 *
 * Even though the int3 trap handler doesn't use the IST anymore in
 * the native case since upstream commit d8ba61ba58c8 ("x86/entry/64:
 * Don't use IST entry for #BP stack"), XEN still treats it as if it
 * would and registers a substitute, xen_int3(), with the hypervisor
 * instead. Until recently, int3() and xen_int3() had identical
 * implementations and the XEN PV code substituting one for another
 * had simply been superfluous. Then, int3_emulate_call() support was
 * implemented and the the int3 entry code is now expected to arrange
 * for a gap on the stack. Unfortunately, the XEN PV int3 handler
 * implementation had been missed and, in consequence,
 * ftrace_int3_handler() can make XEN PV kernels crash. This affects
 * live patching, as ftrace_int3_handler() can be executed during
 * patch application (if one of the patched functions happens to be
 * invoked at certain points during the transition).
 *
 * This shall be fixed from the live patch's module_init(), i.e.
 * before the live patch itself gets applied.
 *
 * The simplest approach would be to just issue a hypercall and
 * register the native int3 handler with the XEN hypervisor on each
 * CPU. However, this is not guaranteed to last forever, because any
 * load_idt() invocation would revert this. One example when this
 * happens is the loading of the tracing IDT, another is loading the
 * "debug IDT" when an NMI finds itself interrupting the debug
 * exception handler. Additionally live patching xen_load_idt() to not
 * replace the int3 handler with the XEN PV implementation would not
 * work reliably during the transition. In particular not for the
 * latter case -- doing that would on the contrary only increase the
 * probability of failure.
 *
 * Thus, the approach chosen here is to rewrite all IDTs' int3 entries
 * with an address to some stub jmp'ing directly to the native int3()
 * implementation. By doing so, xen_load_idt() will subsequently fail
 * to identify the IDT entries' target addresses with int3() and pass
 * them to the hypervisor unmodified.
 */

#if IS_ENABLED(CONFIG_X86_64)

#if !IS_ENABLED(CONFIG_DEBUG_SET_MODULE_RONX)
#error "Live patch supports only CONFIG_DEBUG_SET_MODULE_RONX=y"
#endif

#if !IS_ENABLED(CONFIG_PAGE_TABLE_ISOLATION)
#error "Live patch supports only CONFIG_PAGE_TABLE_ISOLATION=y"
#endif

#include <linux/mm.h>
#include <linux/string.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>
#include <linux/cpu.h>
#include <asm/cacheflush.h>
#include <asm/desc.h>
#include <asm/traps.h>
#include <asm/fixmap.h>
#include <xen/events.h>
#include <xen/xen.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kgraft.h>
#include "kgr_patch_bsc1157770.h"
#include "../kallsyms_relocs.h"
#include "../shadow.h"


static void *(*kgre_module_alloc)(unsigned long size);
static void (*kgre_module_memfree)(void *module_region);
static asmlinkage void (*kgre_int3)(void);
static int (*kgre_set_memory_ro)(unsigned long addr, int numpages);
static int (*kgre_set_memory_rw)(unsigned long addr, int numpages);
extern char kgr_xen_int3_stub_templ_begin[];
extern char kgr_xen_int3_stub_templ_end[];
extern char kgr_xen_int3_stub_templ_jmp_reloc[];

static void *kgr_prepare_xen_int3_stub(void)
{
	void *stub_text;
	unsigned long stub_size, stub_size_aligned;
	char *reloc_loc;
	unsigned long reloc_addr;
	int r;

	/*
	 * Prepare an int3 entry stub to be referenced from the XEN
	 * IDT: copy the stub template to a page and apply the
	 * relocation to the original, non-xen int3 entry handler.
	 * Add KPTI mappings.
	 */
	stub_size = kgr_xen_int3_stub_templ_end - kgr_xen_int3_stub_templ_begin;
	stub_size_aligned = PAGE_ALIGN(stub_size);

	stub_text = kgre_module_alloc(stub_size_aligned);
	if (!stub_text)
		return ERR_PTR(-ENOMEM);

	memzero_explicit(stub_text, stub_size_aligned);
	memcpy(stub_text, kgr_xen_int3_stub_templ_begin, stub_size);

	/*
	 * Apply relocation at the jmp instruction to the original,
	 * non-xen int3 entry code handler.
	 */
	reloc_loc = ((char *)stub_text +
		     (kgr_xen_int3_stub_templ_jmp_reloc -
		      kgr_xen_int3_stub_templ_begin));
	/*
	 * RIP-relative addressing is relative to the next insn,
	 * i.e. relative to reloc_loc + sizof(int).
	 */
	reloc_addr = (unsigned long)reloc_loc + 4;
	reloc_addr = (unsigned long)kgre_int3 - reloc_addr;

	/*
	 * disp32 gets sign-extended. Check that all 32 high bits
	 * match the sign bit, i.e. bit 31. This sanity check should
	 * not trigger, because the module_alloc() from above should
	 * guarantee that stub_text is not too far away from kernel
	 * .text.
	 */
	if (((reloc_addr & (1UL << 31)) && (reloc_addr >> 32 != ~0U)) ||
		(!(reloc_addr & (1UL << 31)) && (reloc_addr >> 32 != 0U))) {
		pr_err("kgraft-patch: relative displacement exceeds 32 bits\n");
		kgre_module_memfree(stub_text);
		return ERR_PTR(-EOVERFLOW);
	}

	WRITE_ONCE(*(unsigned int *)reloc_loc, (unsigned int)reloc_addr);
	flush_icache_range((unsigned long)stub_text,
			   (unsigned long)stub_text + stub_size_aligned);

	/*
	 * Remap the entry stub as RX, c.f. set_section_ro_nx() from
	 * the kernel/module.c
	 */
	r = kgre_set_memory_ro((unsigned long)stub_text,
			       stub_size_aligned >> PAGE_SHIFT);
	if (r) {
		kgre_module_memfree(stub_text);
		return ERR_PTR(r);
	}

	/*
	 * Note: KAISER is ineffective on XEN PV guests,
	 * c.f. kaiser_check_boottime_disable()
	 */

	return stub_text;
}

static void kgr_free_xen_int3_stub(void *stub_text)
{
	unsigned long stub_size;
	int r;

	stub_size = kgr_xen_int3_stub_templ_end - kgr_xen_int3_stub_templ_begin;
	stub_size = PAGE_ALIGN(stub_size);

	r = kgre_set_memory_rw((unsigned long)stub_text,
			       stub_size >> PAGE_SHIFT);
	if (r) {
		/*
		 * This is unlikey to happen. If it does anyway, leak
		 * the memory.
		 */
		return;
	}
	kgre_module_memfree(stub_text);
}


/* from arch/x86/include/asm/desc.h */
static u32 __percpu (*kgre_debug_idt_ctr);
static struct desc_ptr (*kgre_debug_idt_descr);

static inline bool kgrr_is_debug_idt_enabled(void)
{
	if (READ_ONCE(*this_cpu_ptr(kgre_debug_idt_ctr)))
		return true;

	return false;
}

static inline void kgrr_load_debug_idt(void)
{
	load_idt((const struct desc_ptr *)&(*kgre_debug_idt_descr));
}


static atomic_t (*kgre_trace_idt_ctr);
static struct desc_ptr (*kgre_trace_idt_descr);

static inline bool kgrr_is_trace_idt_enabled(void)
{
	if (atomic_read(&(*kgre_trace_idt_ctr)))
		return true;

	return false;
}

static inline void kgrr_load_trace_idt(void)
{
	load_idt((const struct desc_ptr *)&(*kgre_trace_idt_descr));
}

static struct desc_ptr (*kgre_idt_descr);

static inline void kgrr_load_current_idt(void)
{
	if (kgrr_is_debug_idt_enabled())
		kgrr_load_debug_idt();
	else if (kgrr_is_trace_idt_enabled())
		kgrr_load_trace_idt();
	else
		load_idt((const struct desc_ptr *)&(*kgre_idt_descr));
}

static void kgr_load_current_idt_cpu_func(struct work_struct *w)
{
	kgrr_load_current_idt();
}

static int (*kgre_schedule_on_each_cpu)(work_func_t func);

/* Fail-safe implementation of schedule_on_each_cpu(). */
static void kgr_schedule_on_each_cpu_nofail(work_func_t func)
{
	int r;
	int cpu;
	struct work_struct fallback_work;

	/*
	 * schedule_on_each_cpu() can fail only with the per-CPU work
	 * struct_allocation. So it's all or nothing.
	 */
	r = kgre_schedule_on_each_cpu(func);
	if (likely(!r))
		return;

	/*
	 * It failed, process each online CPU individually by means of
	 * the on-stack fallback_work.
	 */
	get_online_cpus();
	for_each_online_cpu(cpu) {
		struct work_struct *work = &fallback_work;

		INIT_WORK(work, func);
		schedule_work_on(cpu, work);
		flush_work(work);
	}
	put_online_cpus();
}

static gate_desc *kgr_tmp_idt_table;

static gate_desc (*kgre_idt_table);

static void kgr_rewrite_idt_entry_addr(struct desc_ptr *idt_descr,
				       int entry,
				       unsigned long new_addr)
{
	gate_desc *orig_idt_table;
	gate_desc gate;

	/*
	 * Due to the struct gate_desc format, an IDT entry's handler
	 * address can't be changed atomically.
	 *
	 * With XEN PV, the IDT table won't be used by HW directly,
	 * but xen_load_idt() can sweep through it concurrently at any
	 * time. Thus, install a temporary copy table while modifying
	 * the original.
	 */
	orig_idt_table = (gate_desc *)idt_descr->address;

	/*
	 * The primary IDT table's virtual address is from a read-only
	 * fixmap. Special case and substitute the known writable
	 * mapping for it.
	 */
	if ((unsigned long)orig_idt_table == fix_to_virt(FIX_RO_IDT))
		orig_idt_table = kgre_idt_table;

	memcpy(kgr_tmp_idt_table, orig_idt_table,
	       NR_VECTORS * sizeof(gate_desc));
	/*
	 * xchg is atomic even on non-aligned destination addresses,
	 * including locations crossing cacheline or page boundaries.
	 */
	smp_wmb();
	xchg(&idt_descr->address, (unsigned long)kgr_tmp_idt_table);
	/*
	 * Note: with XEN PV, the HW is not accessing the IDT table
	 * directly and it should suffice to synchronize with the XEN
	 * PV guest code. Thus, scheduling an "empty" work func here
	 * should be enough in principle. But better be safe than
	 * sorry.
	 */
	kgr_schedule_on_each_cpu_nofail(kgr_load_current_idt_cpu_func);

	/*
	 * At this point, nothing is using orig_idt_table and it's
	 * safe to update it non-atomically. Repack the gate entry
	 * with the new_addr.
	 */
	memcpy(&gate, &orig_idt_table[entry], sizeof(gate));
	pack_gate(&gate, gate.type, new_addr, gate.dpl, gate.ist, gate.segment);
	memcpy(&orig_idt_table[entry], &gate, sizeof(gate));

	/*
	 * Install the original IDT table back. The xen_load_idt()
	 * executed on each CPU will properly announce the changed
	 * entry to XEN.
	 */
	smp_wmb();
	xchg(&idt_descr->address, (unsigned long)orig_idt_table);
	kgr_schedule_on_each_cpu_nofail(kgr_load_current_idt_cpu_func);

	memzero_explicit(kgr_tmp_idt_table, NR_VECTORS * sizeof(gate_desc));
}

static int kgr_prepare_tmp_idt_table(void)
{
	unsigned long size_aligned;
	size_aligned = PAGE_ALIGN(NR_VECTORS * sizeof(gate_desc));
	kgr_tmp_idt_table = vzalloc(size_aligned);
	if (!kgr_tmp_idt_table)
		return -ENOMEM;

	return 0;
}

static void kgr_free_tmp_idt_table(void)
{
	vfree(kgr_tmp_idt_table);
	kgr_tmp_idt_table = NULL;
}

static int kgr_verify_orig_idt_int3(struct desc_ptr *idt_descr)
{
	gate_desc *table;
	gate_desc int3_gate;

	/*
	 * Verify that all X86_TRAP_BP entries point to the original
	 * xen_int3() and haven't got ->ist set.
	 */
	table = (gate_desc *)idt_descr->address;
	memcpy(&int3_gate, &table[X86_TRAP_BP], sizeof(int3_gate));
	if (gate_offset(int3_gate) != (unsigned long)kgre_int3) {
		pr_err("kgraft-patch: unexpected address in int3 IDT entry");
		return -EINVAL;
	}
	if (int3_gate.ist) {
		pr_err("kgraft-patch: unexpected IST value in int3 IDT entry");
		return -EINVAL;
	}

	return 0;
}

static int kgr_install_xen_int3_replacement(void *xen_int3_stub)
{
	int r;

	r = kgr_verify_orig_idt_int3(kgre_idt_descr);
	if (r)
		return r;
	r = kgr_verify_orig_idt_int3(kgre_debug_idt_descr);
	if (r)
		return r;
	r = kgr_verify_orig_idt_int3(kgre_trace_idt_descr);
	if (r)
		return r;

	kgr_rewrite_idt_entry_addr(kgre_idt_descr, X86_TRAP_BP,
				   (unsigned long)xen_int3_stub);
	kgr_rewrite_idt_entry_addr(kgre_debug_idt_descr, X86_TRAP_BP,
				   (unsigned long)xen_int3_stub);
	kgr_rewrite_idt_entry_addr(kgre_trace_idt_descr, X86_TRAP_BP,
				   (unsigned long)xen_int3_stub);

	return 0;
}

static void kgr_uninstall_xen_int3_replacement(void)
{
	kgr_rewrite_idt_entry_addr(kgre_idt_descr, X86_TRAP_BP,
				   (unsigned long)kgre_int3);
	kgr_rewrite_idt_entry_addr(kgre_debug_idt_descr, X86_TRAP_BP,
				   (unsigned long)kgre_int3);
	kgr_rewrite_idt_entry_addr(kgre_trace_idt_descr, X86_TRAP_BP,
				   (unsigned long)kgre_int3);
	/*
	 * The schedule_on_each_cpu() invocation at the very end of
	 * kgr_rewrite_idt_entry_addr() will guarantee that no task
	 * has got its $RIP in the stub entry.
	 */
}


static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "module_alloc", (void *)&kgre_module_alloc },
	{ "module_memfree", (void *)&kgre_module_memfree },
	{ "int3", (void *)&kgre_int3 },
	{ "set_memory_ro", (void *)&kgre_set_memory_ro },
	{ "set_memory_rw", (void *)&kgre_set_memory_rw },
	{ "debug_idt_ctr", (void *)&kgre_debug_idt_ctr },
	{ "debug_idt_descr", (void *)&kgre_debug_idt_descr },
	{ "trace_idt_ctr", (void *)&kgre_trace_idt_ctr },
	{ "trace_idt_descr", (void *)&kgre_trace_idt_descr },
	{ "idt_descr", (void *)&kgre_idt_descr },
	{ "schedule_on_each_cpu", (void *)&kgre_schedule_on_each_cpu },
	{ "idt_table", (void *)&kgre_idt_table },
};

#define KGR_BSC1157770_SHARED_STATE_ID KGR_SHADOW_ID(1157770, 0)

/* Protected by module_mutex. */
struct kgr_bsc1157770_shared_state
{
	unsigned long refcount;
	void *int3_stub_text;
	unsigned long reserved; /* in case the protocol needs to get extended */
};

struct kgr_bsc1157770_shared_state *kgr_bsc1157770_shared_state;

static int kgr_bsc1157770_init_shared_state(void *obj,
					    void *shadow_data,
					    void *ctor_dat)
{
	memset(shadow_data, 0, sizeof(struct kgr_bsc1157770_shared_state));
	return 0;
}

int kgr_patch_bsc1157770_init(void)
{
	int r;
	void *xen_int3_stub;

	/*
	 * The broken xen_int3() has only been installed for PV guests
	 * (but not for PVH guests), c.f. xen_start_kernel, the entry
	 * point for PV + PVH guests.
	 */
	if (!xen_pv_domain() || xen_pvh_domain())
		return 0;

	r = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	if (r)
		return r;

	xen_int3_stub = kgr_prepare_xen_int3_stub();
	if (IS_ERR(xen_int3_stub))
		return PTR_ERR(xen_int3_stub);

	r = kgr_prepare_tmp_idt_table();
	if (r) {
		kgr_free_xen_int3_stub(xen_int3_stub);
		return r;
	}

	mutex_lock(&module_mutex);
	kgr_bsc1157770_shared_state =
		kgr_shadow_get_or_alloc(NULL, KGR_BSC1157770_SHARED_STATE_ID,
					sizeof(*kgr_bsc1157770_shared_state),
					GFP_KERNEL,
					kgr_bsc1157770_init_shared_state, NULL);
	if (!kgr_bsc1157770_shared_state) {
		mutex_unlock(&module_mutex);
		kgr_free_tmp_idt_table();
		kgr_free_xen_int3_stub(xen_int3_stub);
		return -ENOMEM;
	}

	if (!kgr_bsc1157770_shared_state->refcount) {
		/*
		 * We're first, install the xen_int3 stub replacement.
		 */
		r = kgr_install_xen_int3_replacement(xen_int3_stub);
		if (r) {
			kgr_shadow_free(NULL, KGR_BSC1157770_SHARED_STATE_ID,
					NULL);
			mutex_unlock(&module_mutex);
			kgr_free_tmp_idt_table();
			kgr_free_xen_int3_stub(xen_int3_stub);
			return r;
		}
		kgr_bsc1157770_shared_state->int3_stub_text = xen_int3_stub;
		xen_int3_stub = NULL;
	}
	++kgr_bsc1157770_shared_state->refcount;
	mutex_unlock(&module_mutex);

	if (xen_int3_stub) {
		/*
		 * Our int3 stub had not been used. Keep
		 * kgr_tmp_idt_table though, it might be needed again
		 * from *_cleanup().
		 */
		kgr_free_xen_int3_stub(xen_int3_stub);
	}

	return 0;
}

void kgr_patch_bsc1157770_cleanup(void)
{
	void *xen_int3_stub = NULL;

	/* C.f. kgr_patch_bsc1157770_init() */
	if (!xen_pv_domain() || xen_pvh_domain())
		return;

	mutex_lock(&module_mutex);
	--kgr_bsc1157770_shared_state->refcount;
	if (!kgr_bsc1157770_shared_state->refcount) {
		kgr_uninstall_xen_int3_replacement();
		xen_int3_stub = kgr_bsc1157770_shared_state->int3_stub_text;
		kgr_shadow_free(NULL, KGR_BSC1157770_SHARED_STATE_ID, NULL);
	}
	mutex_unlock(&module_mutex);
	kgr_free_tmp_idt_table();
	if (xen_int3_stub)
		kgr_free_xen_int3_stub(xen_int3_stub);
}

#endif /* IS_ENABLED(CONFIG_X86_64) */
